/**
 *
 * fil - Foddex Image Library
 *
 * Copyright (c) 2010 Marc "Foddex" Oude Kotte
 *
 */

#include "fil_internal.h"
#include "ctype.h"
#include <algorithm>
#include <map>
#ifdef _MSC_VER
#pragma warning ( disable : 4996 )	// unsafe STDLIB functions
#endif

//-----------------------------------------------------------------------------
// static functions
//-----------------------------------------------------------------------------
unsigned char fil_get_component_count( fil_pixel_format format ) {

	switch (format) {

		case FIL_RGB:
		case FIL_BGR:
			return 3;

		case FIL_RGBA:
		case FIL_BGRA:
			return 4;

		default:
			return 0;
	}
}

//-----------------------------------------------------------------------------
void fil_init_bmp();
void fil_init_jpeg();
void fil_init_tiff();
void fil_init_tga();
void fil_init_png();
static bool fil_initialized = false;
void fil_initialize() {

	if (fil_initialized)
		return;
	fil_initialized = true;

	fil_init_bmp();
	fil_init_jpeg();
	fil_init_tiff();
	fil_init_tga();
	fil_init_png();
}


//-----------------------------------------------------------------------------
// fil_pixels
//-----------------------------------------------------------------------------
fil_pixels::fil_pixels()
:	mWidth( 0 )
,	mHeight( 0 )
,	mPixelData( 0 )
,	mFormat( FIL_UNKNOWN ) {

}

//-----------------------------------------------------------------------------
fil_pixels::~fil_pixels() {

	cleanup();
}

//-----------------------------------------------------------------------------
void fil_pixels::cleanup() {

	if (mPixelData) {
		delete [] mPixelData;
		mPixelData = 0;
	}
	mWidth = mHeight = 0;
	mFormat = FIL_UNKNOWN;
	mFilename.clear();
}

//-----------------------------------------------------------------------------
void fil_pixels::initialize( const std::string& filename, int width, int height, unsigned char* pixel_data, fil_pixel_format format ) {

	cleanup();
	mFilename = filename;
	mWidth = width;
	mHeight = height;
	mPixelData = pixel_data;
	mFormat = format;
}

//-----------------------------------------------------------------------------
int fil_pixels::get_width() const {

	return mWidth;
}

//-----------------------------------------------------------------------------
int fil_pixels::get_height() const {

	return mHeight;
}

//-----------------------------------------------------------------------------
unsigned char* fil_pixels::get_pixel_data() const {

	return mPixelData;
}

//-----------------------------------------------------------------------------
unsigned char fil_pixels::get_bytes_per_pixel() const {

	return fil_get_component_count( mFormat );
}

//-----------------------------------------------------------------------------
fil_pixel_format fil_pixels::get_format() const {

	return mFormat;
}

//-----------------------------------------------------------------------------
// global, exported functions
//-----------------------------------------------------------------------------
static std::vector<std::string>*			_fil_formats = NULL;			// only used to quickly return to the user a list of supported formats
static std::map<std::string, fil_loader>*	_fil_format_map = NULL;			// the internal format map

//-----------------------------------------------------------------------------
const std::vector<std::string>& fil_get_supported_formats() {

	fil_initialize();
	return *_fil_formats;
}

//-----------------------------------------------------------------------------
fil_load_result fil_load_from_file( const std::string& filename, fil_pixels* target, unsigned int ppflags ) {

	fil_initialize();

	// check input
	if (!target)
		return FIL_NULL_INPUT;

	// attempt to open the file and determine its size
	FILE* f = fopen( filename.c_str(), "rb" );
	if (!f)
		return FIL_FILE_OPEN_ERROR;
	fseek( f, 0, SEEK_END );
	long size = ftell( f );
	fseek( f, 0, SEEK_SET );
	
	// attempt to read the raw file
	char* memory = new char[size];
	if (!memory)
		return FIL_MEMORY_ALLOCATION_ERROR;
	if (fread( memory, 1, size, f )!=static_cast<size_t>( size )) {
		fclose( f );
		delete [] memory;
		return FIL_FILE_READ_ERROR;
	}
	
	// load using memory loader, and cleanup
	fil_load_result res = fil_load_from_memory( memory, size, filename, target, ppflags );
	delete[] memory;
	return res;
}

//-----------------------------------------------------------------------------
fil_load_result fil_load_from_memory( char* image_data, size_t bytes, const std::string& format, fil_pixels* target, unsigned int ppflags ) {

	fil_initialize();

	// check input
	if (!image_data || !target)
		return FIL_NULL_INPUT;

	// parse the format from the format string
	size_t offset = format.find_last_of( '.' );
	if (offset==std::string::npos)
		offset = 0;
	else
		offset++;
	std::string ext = format.empty() ? "" : format.substr( offset );
	for (size_t i=0; i<ext.size(); ++i)
		ext[i] = tolower( ext[i] );

	// check if we support this!
	std::map<std::string, fil_loader>::const_iterator it = _fil_format_map->find( ext );
	if (it==_fil_format_map->end())
		return FIL_UNSUPPORTED_FORMAT;
	fil_data_reader r( image_data, bytes, format );
	return (*it).second( r, *target, ppflags );
}

//-----------------------------------------------------------------------------
// internal functions
//-----------------------------------------------------------------------------
static void fil_cleanup_format_data() {
	delete _fil_format_map;
	delete _fil_formats;
}

//-----------------------------------------------------------------------------
void fil_register_format( const std::string& format, fil_loader loader ) {

	if (!_fil_format_map) {
		_fil_format_map = new std::map<std::string, fil_loader>;
		_fil_formats = new std::vector<std::string>;
		atexit( &fil_cleanup_format_data );
	}

	// do not check loader, do not check if format is already used, do not lower case, 
	// we're an internal function and (hopefully) the author isn't crazy ^^
	(*_fil_format_map)[format] = loader;
	(*_fil_formats).push_back( format );
}

//-----------------------------------------------------------------------------
fil_data_reader::fil_data_reader( const char* data, size_t size, const std::string& filename )
:	mData( data )
,	mSize( size )
,	mPosition( 0 )
,	mError( false )
,	mFilename( filename ) {

}

//-----------------------------------------------------------------------------
void fil_data_reader::skip( size_t bytes ) {

	if (mError)
		return;
	if (mPosition + bytes > mSize)
		mError = true;
	else
		mPosition += bytes;
}

//-----------------------------------------------------------------------------
void fil_data_reader::seek( size_t bytes ) {

	if (mError)
		return;
	if (mSize < bytes)
		mError = true;
	else
		mPosition = bytes;
}

//-----------------------------------------------------------------------------
bool fil_data_reader::in_error_state() {

	return mError;
}

//-----------------------------------------------------------------------------
void fil_data_reader::read( void* target, size_t bytes ) {

	if (mError) {
		memset( target, 0, bytes );
		return;
	}
	if (mPosition + bytes > mSize) {
		mError = true;
		memset( target, 0, bytes );
	} else {
		memcpy( target, mData + mPosition, bytes );
		mPosition += bytes;
	}
}

//-----------------------------------------------------------------------------
const std::string& fil_data_reader::get_filename() const {

	return mFilename;
}

//-----------------------------------------------------------------------------
const char* fil_data_reader::get_data() const {

	return mData;
}

//-----------------------------------------------------------------------------
size_t fil_data_reader::get_size() const {

	return mSize;
}

//-----------------------------------------------------------------------------
/* static */ void fil_tools::install_image( fil_pixels& target, const std::string& filename, int width, int height, unsigned char* pixel_data, fil_pixel_format format, unsigned int ppflags ) {

	// perform postprocessing!
	if (ppflags & FIL_PP_SETCOLORLINEARALPHA) {
		switch (format) {
            case FIL_RGB:
			case FIL_BGR:
            {
                // make material RGBA
				unsigned char* newpixels = new unsigned char [4 * (unsigned int)(width) * (unsigned int)(height) ];
                unsigned char* oldpixels = pixel_data;
                for (unsigned int i=0; i<(unsigned int)(height)*(unsigned int)(width); ++i) {
                        newpixels[0] = oldpixels[0];
                        newpixels[1] = oldpixels[1];
                        newpixels[2] = oldpixels[2];
                        float alpha = (newpixels[0] + newpixels[1] + newpixels[2]) / 3.f;
                        newpixels[3] = (unsigned char)( alpha );
                        newpixels += 4;
                        oldpixels += 3;
                }
				delete [] pixel_data;
                pixel_data = newpixels;
				format = format==FIL_RGB ? FIL_RGBA : FIL_BGRA;
                break;
            }

            case FIL_RGBA:
			case FIL_BGRA:
            {
                unsigned char* pixels = pixel_data;
                for (int i=0; i<height*width; ++i) {
                        float alpha = (pixels[0] + pixels[1] + pixels[2]) / 3.f;
                        pixels[3] = (unsigned char)( alpha );
                        pixels += 4;
                }
                break;
            }
			default:
				break;
		}
	}

	if (ppflags & FIL_PP_SETALPHAWHERENOTBLACK) {
		switch (format) {
            case FIL_RGB:
			case FIL_BGR:
            {
                // make material RGBA
				unsigned char* newpixels = new unsigned char [4 * (unsigned int)(width) * (unsigned int)(height) ];
                unsigned char* oldpixels = pixel_data;
                for (unsigned int i=0; i<(unsigned int)(height)*(unsigned int)(width); ++i) {
                        newpixels[0] = oldpixels[0];
                        newpixels[1] = oldpixels[1];
                        newpixels[2] = oldpixels[2];
						newpixels[3] = (newpixels[0] + newpixels[1] + newpixels[2])>0 ? 255 : 0;
                        newpixels += 4;
                        oldpixels += 3;
                }
				delete [] pixel_data;
                pixel_data = newpixels;
				format = format==FIL_RGB ? FIL_RGBA : FIL_BGRA;
                break;
            }

            case FIL_RGBA:
			case FIL_BGRA:
            {
                unsigned char* pixels = pixel_data;
                for (int i=0; i<height*width; ++i) {
						pixels[3] = (pixels[0] + pixels[1] + pixels[2]) > 0 ? 255 : 0;
                        pixels += 4;
                }
                break;
            }
			default:
				break;
		}
	}

	// initialize!
	target.initialize( filename, width, height, pixel_data, format );
}

//-----------------------------------------------------------------------------
/* static */ void fil_tools::flip_vertical( int width, int height, unsigned char* pixel_data, fil_pixel_format format ) {

	int componentCount = fil_get_component_count( format );
	int rowByteSize = width * componentCount;
	int swapRowCount = height>>1;
	unsigned char* swapRow = new unsigned char[rowByteSize];
	unsigned char* sourceRow = pixel_data;
	unsigned char* targetRow = sourceRow + (int(height)-1) * rowByteSize;
	for (int y=0; y<swapRowCount; ++y) {
		memcpy( swapRow, sourceRow, rowByteSize );
		memcpy( sourceRow, targetRow, rowByteSize );
		memcpy( targetRow, swapRow, rowByteSize );
		sourceRow += rowByteSize;
		targetRow -= rowByteSize;
	}
	delete [] swapRow;
}
