/**
 *
 * fil - Foddex Image Library
 *
 * Copyright (c) 2010 Marc "Foddex" Oude Kotte
 *
 */

#ifndef __FIL_H__
#define __FIL_H__

#include <string.h>
#include <stdio.h>
#include <string>
#include <vector>

#if !defined(FIL_STATIC) && defined(_MSC_VER)
#	ifdef FIL_EXPORTS
#		define	FIL_API		__declspec(dllexport)
#	else
#		define FIL_API		__declspec(dllimport)
#	endif
#else
#	define FIL_API
#endif

/** enumeration of pixel formats as returned in data */
enum fil_pixel_format {

	FIL_UNKNOWN = 0,
	FIL_RGB,
	FIL_BGR,
	FIL_RGBA,
	FIL_BGRA
};

/** possible load results */
enum fil_load_result {

	// this is the only 'success' result
	FIL_OK = 0,

	// here come the error results
	FIL_NULL_INPUT,						// one or more mandatory input values was NULL
	FIL_FILE_OPEN_ERROR,				// failed to open specified file
	FIL_FILE_READ_ERROR,				// failed to read specified file
	FIL_MEMORY_ALLOCATION_ERROR,		// failed to allocate required memory
	FIL_COULDNT_DETERMINE_FORMAT,		// failed to obtain format from filename
	FIL_UNSUPPORTED_FORMAT,				// failed to load image, extension does not indicate a valid format
	FIL_FORMAT_ERROR,					// failed to handle image based on extension
	FIL_UNSUPPORTED_SUB_FORMAT,			// failed to handle image, subformat of format not supported (e.g. grayscale BMP, only 24-bits supported)
	FIL_GENERIC_ERROR					// generic something-went-wrong-error
};

/** post process flags, set 0 to have no post processing done */
enum fil_post_process_flags {

	FIL_PP_SETCOLORLINEARALPHA		= 0x01,		// makes textures always 32-bits, and sets per-pixel alpha value linear to total RGB pixel color
	FIL_PP_SETALPHAWHERENOTBLACK	= 0x02		// makes textures always 32-bits, and sets alpha to 255 for non-pitch-black values, and 0 for pitch-black values
};

/** main pixel data class; cleans up its own data in the constructor */
class FIL_API fil_pixels {

	friend struct fil_tools;

	private:
		int					mWidth;
		int					mHeight;
		unsigned char*		mPixelData;
		fil_pixel_format	mFormat;
		std::string			mFilename;
		void cleanup();

		// not implemented
		fil_pixels( const fil_pixels& rhs );
		fil_pixels& operator=( const fil_pixels& rhs );

		// only accessible from within fil library
		void initialize( const std::string& filename, int width, int height, unsigned char* pixel_data, fil_pixel_format format );

	public:
		fil_pixels();
		~fil_pixels();

		int get_width() const;
		int get_height() const;
		unsigned char* get_pixel_data() const;
		unsigned char get_bytes_per_pixel() const;
		fil_pixel_format get_format() const;
};

/** Returns the list of supported formats. Returns a vector of strings, containing all recognized and supported
 *	file extensions.
 */
FIL_API const std::vector<std::string>& fil_get_supported_formats();

/**
 * Generic file loader, loads an image from a file, and fills a fil_pixels structure.
 * Allocate a target fil_pixels object yourself, and then specify its pointer as 'target' to
 * this function. Returns FIL_OK upon success, any of the other fil_load_result enumeration
 * items upon error.
 *
 * note: image format is recognized by filename extension;
 */
FIL_API fil_load_result fil_load_from_file( const std::string& filename, fil_pixels* target, unsigned int ppflags );

/**
 * Generic memory loader, loads an image from a raw file that was copied into memory.
 * Allocate a target fil_pixels object yourself, and then specify its pointer as 'target' to
 * this function. Returns FIL_OK when image was succesfully loaded, returns other value upon error.
 *
 * The specified format string may not be empty, the library is not going to "guess" what the format
 * is based on the image_data. If there's one or more .'s in the format, everything up to and including
 * the last dot is ignored. This way you can specify the original filename as the format safely.
 */
FIL_API fil_load_result fil_load_from_memory( char* image_data, size_t bytes, const std::string& format, fil_pixels* target, unsigned int ppflags );

#endif
