/**
 *
 * fil - Foddex Image Library
 *
 * Copyright (c) 2010 Marc "Foddex" Oude Kotte
 *
 */

#include "fil_internal.h"

//--------------------------------------------------------------------------------
// http://partners.adobe.com/public/developer/en/tiff/TIFF6.pdf

struct tiffIFD {
	unsigned short Tag;
	unsigned short FieldType;
	unsigned int FieldCount;
	unsigned int ValueOrOffset;
}; //sizeof(tiffIFD)==12!!

enum tiffFieldType {
	TFT_BYTE = 1,
	TFT_ASCII = 2,
	TFT_SHORT = 3,
	TFT_LONG = 4,
	TFT_RATIONAL = 5,
	TFT_SBYTE = 6,
	TFT_UNDEFINED = 7,
	TFT_SSHORT = 8,
	TFT_SLONG = 9,
	TFT_SRATIONAL = 10,
	TFT_FLOAT = 11,
	TFT_DOUBLE = 12
};

//--------------------------------------------------------------------------------
static const tiffIFD* FindIFD( const std::vector<tiffIFD>& ifds, unsigned short tag ) {
	for (size_t i=0; i<ifds.size(); ++i) {
		if (ifds[i].Tag==tag) return &ifds[i];
	}
	return NULL;
}

//--------------------------------------------------------------------------------
static unsigned int GetUINT32Value( const tiffIFD& ifd ) {

	switch (ifd.FieldType) {
		case TFT_BYTE: 
		case TFT_SBYTE:
			return ifd.ValueOrOffset & 0xFF;
		case TFT_SHORT: 
		case TFT_SSHORT:
			return ifd.ValueOrOffset & 0xFFFF;
		case TFT_LONG: 
		case TFT_SLONG:
			return ifd.ValueOrOffset & 0xFFFFFFFF;
	}
	return 0;
}

//--------------------------------------------------------------------------------
static fil_load_result load_tiff( fil_data_reader& reader, fil_pixels& target, unsigned int ppflags ) {

	// check endianness, only support little endian atm
	unsigned short endianness;
	reader( endianness );
	if (endianness!=0x4949) 
		return FIL_UNSUPPORTED_SUB_FORMAT;

	// check tiff indentification
	unsigned short tiffIdent;
	reader( tiffIdent );
	if (tiffIdent!=42) 
		return FIL_FORMAT_ERROR;

	// get offset of first IFD
	unsigned int firstIFD;
	reader( firstIFD );

	// get IFD's
	std::vector<tiffIFD> ifds;
	unsigned int nextOffset = firstIFD;
	reader.seek( nextOffset );
	if (reader.in_error_state())
		return FIL_FORMAT_ERROR;
	unsigned short ifdCount;
	reader( ifdCount );
	for (unsigned short i=0; i<ifdCount; ++i) {
		tiffIFD ifd;
		reader.read( &ifd, 12 );
		ifds.push_back( ifd );
	}
	reader( nextOffset );
	if (nextOffset) 
		return FIL_UNSUPPORTED_SUB_FORMAT;	// multiple images within single file, we don't support that!

	// we support only RGB Full Color images at the moment, find the appropriate tags
	const tiffIFD* tagSamplesPerPixel = FindIFD( ifds, 277 );
	const tiffIFD* tagImageWidth = FindIFD( ifds, 256 );
	const tiffIFD* tagImageLength = FindIFD( ifds, 257 );
	const tiffIFD* tagImageXRes = FindIFD( ifds, 282 );
	const tiffIFD* tagImageYRes = FindIFD( ifds, 283 );
	const tiffIFD* tagBitsPerSample = FindIFD( ifds, 258 );
	const tiffIFD* tagCompression = FindIFD( ifds, 259 );
	const tiffIFD* tagRowsPerStrip = FindIFD( ifds, 278 );
	const tiffIFD* tagStripOffsets = FindIFD( ifds, 273 );
	const tiffIFD* tagOrientation = FindIFD( ifds, 274 );					// not required
	const tiffIFD* tagPhotometricInterpretation = FindIFD( ifds, 262 );		// not required
	const tiffIFD* tagPlanarConfiguration = FindIFD( ifds, 284 );			// not required
	if (!tagSamplesPerPixel || 
		!tagImageWidth || 
		!tagImageLength || 
		!tagImageXRes || 
		!tagImageYRes || 
		!tagBitsPerSample || 
		!tagCompression || 
		!tagRowsPerStrip ||
		!tagStripOffsets) 
		return FIL_FORMAT_ERROR;

	if (tagOrientation && (GetUINT32Value( *tagOrientation )!=1 && GetUINT32Value( *tagOrientation )!=4)) 
		return FIL_UNSUPPORTED_SUB_FORMAT;
	if (tagPhotometricInterpretation && GetUINT32Value( *tagPhotometricInterpretation )!=2) 
		return FIL_UNSUPPORTED_SUB_FORMAT;
	if (tagPlanarConfiguration && GetUINT32Value( *tagPlanarConfiguration )!=1) 
		return FIL_UNSUPPORTED_SUB_FORMAT;

	// we don't support compressed files
	if (GetUINT32Value(*tagCompression)!=1) 
		return FIL_UNSUPPORTED_SUB_FORMAT;
	if (GetUINT32Value(*tagSamplesPerPixel)!=3 && GetUINT32Value(*tagSamplesPerPixel)!=4) 
		return FIL_UNSUPPORTED_SUB_FORMAT;

	// we don't support files that have anything else than 8 bits per sample
	reader.seek( tagBitsPerSample->ValueOrOffset );
	if (reader.in_error_state()) 
		return FIL_UNSUPPORTED_SUB_FORMAT;
	for (size_t i=0; i<GetUINT32Value(*tagSamplesPerPixel); ++i) {
		unsigned short bits;
		reader( bits );
		if (bits!=8) 
			return FIL_UNSUPPORTED_SUB_FORMAT;
	}

	// calc strip count, and get strip offsets
	unsigned int stripCount = GetUINT32Value( *tagImageLength ) / GetUINT32Value( *tagRowsPerStrip );
	if (GetUINT32Value( *tagImageLength ) % GetUINT32Value( *tagRowsPerStrip )) ++stripCount;
	std::vector<unsigned int> stripOffsets;
	stripOffsets.resize( stripCount );
	if (tagStripOffsets->FieldCount==1) {
		stripOffsets[0] = tagStripOffsets->ValueOrOffset;
	} else {
		reader.seek( tagStripOffsets->ValueOrOffset );
		if (reader.in_error_state()) 
			return FIL_FORMAT_ERROR;
		for (size_t i=0; i<stripCount; ++i) {
			unsigned int& offset = stripOffsets[i];
			reader( offset );
			offset = offset + 1 - 1;
		}
	}

	// get memory
	int width, height, bytepp;
	width = GetUINT32Value( *tagImageWidth );
	height = GetUINT32Value( *tagImageLength );
	bytepp = GetUINT32Value( *tagSamplesPerPixel );
	fil_pixel_format format = bytepp==4 ? FIL_RGBA : FIL_RGB;
	unsigned int size = (unsigned int)(width) * (unsigned int)(height) * bytepp;
	unsigned char* data = new unsigned char[size];
	unsigned char* pixels = data;

	// load strips
	unsigned int rowsLeft = (unsigned int)(height);
	for (size_t i=0; i<stripCount; ++i) {
		reader.seek( (unsigned int)(stripOffsets[i]) );
		if (reader.in_error_state()) 
			return FIL_FORMAT_ERROR;
		unsigned int rowsInStrip = std::min<unsigned int>( rowsLeft, GetUINT32Value( *tagRowsPerStrip ));
		unsigned int byteCount = rowsInStrip * (unsigned int)(width) * GetUINT32Value( *tagSamplesPerPixel );
		reader.read( pixels, byteCount );
		pixels += byteCount;
		rowsLeft -= rowsInStrip;
	}
	if (!( pixels == data + (unsigned int)(width) * (unsigned int)(height) * GetUINT32Value( *tagSamplesPerPixel ) ))
		return FIL_FORMAT_ERROR; // read too much or not enough data, invalid TIFF file?

	// flip vertical if required
	if (tagOrientation && GetUINT32Value( *tagOrientation )==4)
		fil_tools::flip_vertical( width, height, data, format );
	fil_tools::install_image( target, reader.get_filename(), width, height, data, format, ppflags );
	return FIL_OK;
}

/** Automatically register "tiff" as a supported format. */
void fil_init_tiff() {

	fil_format_registrar _fil_tiff_registrar1( "tiff", load_tiff );
	fil_format_registrar _fil_tiff_registrar2( "tif", load_tiff );
}
